// ====================================================
//
// Copyright (c) 2002 Sean Wilson. All Rights Reserved.
//
// ====================================================

#ifdef _WIN32

#include "sandbox.h"
#include "misc.h"
#include "timer.h"
#include "resource.h"

#include <iostream>
#include <sstream>
#include <fstream>

#include <windows.h>
#include <commdlg.h>
#include <new.h>

#ifdef _DEBUG
#include <crtdbg.h>
#endif

namespace
{

/** Application name. */
const char* APPLICATION_NAME = "Sandbox";
/** Window class name. */
const char* WINDOW_CLASS_NAME = "SandboxWindowClassName";
/** Filename buffer size for OPENFILENAME structure. */
const int OPENFILE_FILENAME_SIZE = 1024;

/** File extension for application files. */
const char* APPLICATION_FILE_EXTENSION = "san";
/** File filter for application files. */
const char* APPLICATION_FILE_FILTER = "Sandbox files\0*.san\0\0";

/** Window width. */
const int WINDOW_WIDTH = 800;
/** Window height. */
const int WINDOW_HEIGHT = 600;

/** Maximum frames per second allowable. */
const int MAX_FPS = 150;

/** Application window handle. */
::HWND applicationWindow = 0;
/** Application instance handle. */
::HINSTANCE applicationInstance = 0;
/** Sandbox application. */
std::auto_ptr<seanw::sandbox::Sandbox> application;

/** Whether left mouse button is down. */
bool lmbDown = false;
/** Whether right mouse button is down. */
bool rmbDown = false;
/** Whether cursor is inside window. */
bool mouseInWindow = false;
/** Previous point mouse cursor was recorded. */
::POINT lastMousePoint = {0, 0};

/** Offscreen image for double buffering. */
::HBITMAP offscreen = 0;
/** Offscreen image DC. */
::HDC  offscreenDc = 0;
/** Client area rectangle. */
::RECT clientRect = {0};

/** Currently loaded filename. */
std::string currentFilename;

/** Updates window title to reflect currently loaded filename. */
void updateWindowTitle()
{
	std::ostringstream title;

	title << APPLICATION_NAME;
		
	if (!currentFilename.empty())
	{
		title << " - " << currentFilename;
	}

	SendMessage(applicationWindow, WM_SETTEXT, 0, reinterpret_cast<LPARAM>(title.str().c_str()));	
}

/** Updates mode menu to reflect current application manipulation mode.
 *  @throw std::out_of_range if application is in an invalid mode.
 */
void updateApplcationMode()
{
	int modeMenuId;
	
	switch (application->getMode())
	{
		case seanw::sandbox::Sandbox::MODE_PLAY:  modeMenuId = ID_MENU_MODE_PLAY;  break;
		case seanw::sandbox::Sandbox::MODE_BUILD: modeMenuId = ID_MENU_MODE_BUILD; break;
		case seanw::sandbox::Sandbox::MODE_GLUE:  modeMenuId = ID_MENU_MODE_GLUE;  break;
		default:                                  throw std::out_of_range(seanw::misc::getErrorString(__FILE__, __LINE__, "Bad mode identifier.").c_str());
	}

	::CheckMenuRadioItem(::GetMenu(applicationWindow), ID_MENU_MODE_PLAY, ID_MENU_MODE_GLUE, modeMenuId, MF_BYCOMMAND);
}

/** Updates option menu to reflect current application options. */
void updateApplcationOptions()
{
	::CheckMenuItem(::GetMenu(applicationWindow), ID_MENU_OPTIONS_ENABLE_GRAVITY,  (application->isGravityOn()  ? MF_CHECKED : MF_UNCHECKED) | MF_BYCOMMAND);
	::CheckMenuItem(::GetMenu(applicationWindow), ID_MENU_OPTIONS_ENABLE_FRICTION, (application->isFrictionOn() ? MF_CHECKED : MF_UNCHECKED) | MF_BYCOMMAND);
}

/** Updates window title and menus to reflect current application state.*/
void updateWindowControls()
{
	updateApplcationMode();
	updateWindowTitle();
	updateApplcationOptions();
}

/** Returns true if changes have been made to the application after it was restarted or loaded from a file. */
bool changesMade()
{
	std::ostringstream currentApplicationSaved;

	currentApplicationSaved << *application << "\n";

	std::ostringstream oldApplicationSaved;

	if (currentFilename.empty())
	{
		oldApplicationSaved << seanw::sandbox::Sandbox(clientRect.right, clientRect.bottom) << "\n";
	}
	else
	{
		std::ifstream inputStream(currentFilename.c_str());

		oldApplicationSaved << inputStream.rdbuf();
	}

	return currentApplicationSaved.str() != oldApplicationSaved.str();
}

/** Restarts the application. */
void restartApplication()
{
	application = std::auto_ptr<seanw::sandbox::Sandbox>(new seanw::sandbox::Sandbox(clientRect.right, clientRect.bottom));
	currentFilename = "";
	updateWindowControls();
}

/** Saves application to file.
 *  @param  filename file to save to. If one is not supplied, the user is asked to select one.
 *  @return          true if file was successfully saved.
 */
bool save(std::string filename)
{
	if (filename.empty())
	{
		char           selectedFilename[OPENFILE_FILENAME_SIZE] = "";
		::OPENFILENAME openFilename                             = {0};

		openFilename.lStructSize     = sizeof(::OPENFILENAME);
		openFilename.hwndOwner       = applicationWindow;
		openFilename.lpstrFile       = selectedFilename;
		openFilename.nMaxFile        = sizeof(selectedFilename);
		openFilename.lpstrFilter     = APPLICATION_FILE_FILTER;
		openFilename.nFilterIndex    = 1;
		openFilename.lpstrFileTitle  = 0;
		openFilename.nMaxFileTitle   = 0;
		openFilename.lpstrInitialDir = 0;
		openFilename.lpstrDefExt     = APPLICATION_FILE_EXTENSION;
		openFilename.Flags           = OFN_PATHMUSTEXIST | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;

		// If user did not pick a filename
		if (!::GetSaveFileName(&openFilename))
		{
			return false;
		}

		filename = selectedFilename;
	}	

	try
	{
		std::ofstream outputFile(filename.c_str());

		outputFile.exceptions(std::ios_base::badbit | std::ios_base::failbit | std::ios_base::eofbit);

		outputFile << *application << "\n";

		currentFilename = filename;
		updateWindowControls();
		
		return true;
	}
	catch (...)
	{
		MessageBox(applicationWindow, "File could not be saved.", APPLICATION_NAME, MB_ICONEXCLAMATION);

		return false;
	}
}

/** Gives the user the option to save changes they have made.
 *  @return true if user has saved their changes or have agreed that they can be discarded
 */
bool saveChanges()
{
	if (!changesMade())
	{
		return true;
	}
	
	int buttonPressed = MessageBox(applicationWindow, "Do you want to save the changes you have made?", APPLICATION_NAME, MB_YESNOCANCEL | MB_ICONEXCLAMATION);
		
	if (buttonPressed == IDCANCEL)
	{
		return false;
	}

	if (buttonPressed == IDNO || save(currentFilename))
	{
		return true;
	}

	return false;
}

/** Opens an application file selected by user after giving them the option to save any changes they have made.
 *  @return true if file was successfully opened.
 */
bool open()
{
	if (!saveChanges())
	{
		return false;
	}

	char           selectedFilename[OPENFILE_FILENAME_SIZE] = "";
	::OPENFILENAME openFilename                             = {0};
	
	openFilename.lStructSize     = sizeof(::OPENFILENAME);
	openFilename.hwndOwner       = applicationWindow;
	openFilename.lpstrFile       = selectedFilename;
	openFilename.nMaxFile        = sizeof(selectedFilename);
	openFilename.lpstrFilter     = APPLICATION_FILE_FILTER;
	openFilename.nFilterIndex    = 1;
	openFilename.lpstrFileTitle  = 0;
	openFilename.nMaxFileTitle   = 0;
	openFilename.lpstrInitialDir = 0;
	openFilename.lpstrDefExt     = APPLICATION_FILE_EXTENSION;
	openFilename.Flags           = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;

	// If user did not pick a filename
	if (!::GetOpenFileName(&openFilename))
	{
		return false;
	}
	
	// Another application file will be loaded into first incase current is corrupted by failed load attempt
	std::auto_ptr<seanw::sandbox::Sandbox> newApplication(new seanw::sandbox::Sandbox(clientRect.right, clientRect.bottom));

	try
	{
		std::ifstream inputFile(selectedFilename);

		inputFile.exceptions(std::ios_base::badbit | std::ios_base::failbit | std::ios_base::eofbit);

		inputFile >> *newApplication;

		application     = newApplication;
		currentFilename = selectedFilename;
		
		updateWindowControls();

		return true;
	}
	catch (...)
	{
		MessageBox(applicationWindow, "File could not be loaded.", APPLICATION_NAME, MB_ICONEXCLAMATION);

		return false;
	}
}

/** About window message handler. */
::BOOL CALLBACK aboutDialogProc(::HWND window, ::UINT message, ::WPARAM wParam, ::LPARAM)
{
	switch (message)
	{
		case WM_INITDIALOG:

			::SetDlgItemText(window, IDC_BUILD_TIME, __TIMESTAMP__);

			#ifdef _DEBUG
				::SetDlgItemText(window, IDC_BUILD_TYPE, "Debug build");	
			#else
				::SetDlgItemText(window, IDC_BUILD_TYPE, "Release build");
			#endif

			::SetFocus(window);

			return FALSE;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDC_OK_BUTTON:
					::EndDialog(window, 0);
					return FALSE;
			}

			return FALSE;

		case WM_CLOSE:
			::EndDialog(window, 0);
			return FALSE;
	}

	// Message not processed
	return FALSE;
}

/** Application window message handler. */
::LRESULT CALLBACK windowProcedure(::HWND window, ::UINT message, ::WPARAM wParam, ::LPARAM lParam)
{
	switch (message)
	{	
		case WM_CLOSE:
			if (saveChanges())
			{
				::DestroyWindow(applicationWindow);
			}
			return FALSE;

		case WM_DESTROY:
			::PostQuitMessage(0);
			return FALSE;

		case WM_LBUTTONDOWN:

			application->mouseLeftPressed();
			lmbDown = true;
			
			if (!rmbDown)
			{
				::SetCapture(applicationWindow);
			}

			return FALSE;

		case WM_RBUTTONDOWN:
			
			application->mouseRightPressed();
			rmbDown = true;
			
			if (!lmbDown)
			{
				::SetCapture(applicationWindow);
			}

			return FALSE;

		case WM_LBUTTONUP:

			application->mouseLeftReleased();
			lmbDown = false;

			if (!rmbDown)
			{
				::ReleaseCapture();
			}

			return FALSE;

		case WM_RBUTTONUP:

			application->mouseRightReleased();
			rmbDown = false;
			
			if (!lmbDown)
			{
				::ReleaseCapture();
			}

			return FALSE;

		case WM_CAPTURECHANGED:

			if (lmbDown)
			{
				application->mouseLeftReleased();
				lmbDown = false;
			}

			if (rmbDown)
			{
				application->mouseRightReleased();
				rmbDown = false;
			}
	
			return FALSE;

		case WM_PAINT:
		{
			::PAINTSTRUCT paint;
			::HDC         dc = ::BeginPaint(window, &paint);

			::SelectObject(offscreenDc, offscreen);

			{
				seanw::graphics::Graphics graphics(offscreenDc);

				application->paint(graphics);
			}

			::BitBlt(dc, 0, 0, clientRect.right, clientRect.bottom, offscreenDc, 0, 0, SRCCOPY);
			::EndPaint(window, &paint);

			return FALSE;
		}

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case ID_MENU_FILE_NEW:
					if (saveChanges())
					{
						restartApplication();
					}

					return FALSE;
			
				case ID_MENU_FILE_EXIT:
					::SendMessage(applicationWindow, WM_CLOSE, 0, 0);
					return FALSE;

				case ID_MENU_FILE_OPEN:
					open();
					return FALSE;

				case ID_MENU_FILE_SAVE_AS:
					save("");
					return FALSE;

				case ID_MENU_FILE_SAVE:
					save(currentFilename);
					return FALSE;

				case ID_MENU_OPTIONS_ENABLE_GRAVITY:
					application->setGravity(::GetMenuState(::GetMenu(window), LOWORD(wParam), MF_BYCOMMAND) == MF_UNCHECKED);
					updateWindowControls();
					return FALSE;

				case ID_MENU_OPTIONS_ENABLE_FRICTION:
					application->setFriction(::GetMenuState(::GetMenu(window), LOWORD(wParam), MF_BYCOMMAND) == MF_UNCHECKED);
					updateWindowControls();
					return FALSE;

				case ID_MENU_MODE_PLAY:
					application->setMode(seanw::sandbox::Sandbox::MODE_PLAY);
					updateWindowControls();
					return FALSE;

				case ID_MENU_MODE_BUILD:
					application->setMode(seanw::sandbox::Sandbox::MODE_BUILD);
					updateWindowControls();
					return FALSE;

				case ID_MENU_MODE_GLUE:
					application->setMode(seanw::sandbox::Sandbox::MODE_GLUE);
					updateWindowControls();
					return FALSE;

				case ID_MENU_HELP_ABOUT:
					::DialogBox(applicationInstance, MAKEINTRESOURCE(IDD_ABOUT_DIALOG), applicationWindow, aboutDialogProc);
					return FALSE;
			}
	}

	return ::DefWindowProc(window, message, wParam, lParam);
}

/** Pass on any mouse events to application. */
void updateMouse()
{

	// Rectangle client area uses in screen co-ordinates
	::RECT clientScreenRect = clientRect;
	::GetClientRect(applicationWindow, &clientScreenRect);

	::WINDOWPLACEMENT windowPlacement = {0};
	windowPlacement.length = sizeof(::WINDOWPLACEMENT);
	::GetWindowPlacement(applicationWindow, &windowPlacement);

	::OffsetRect(&clientScreenRect, windowPlacement.rcNormalPosition.left, windowPlacement.rcNormalPosition.top);
	::OffsetRect(&clientScreenRect, ::GetSystemMetrics(SM_CYFIXEDFRAME),
	                                ::GetSystemMetrics(SM_CYMENUSIZE  ) +
	                                ::GetSystemMetrics(SM_CYCAPTION   ) +
	                                ::GetSystemMetrics(SM_CYFIXEDFRAME));

	// Where cursor is in screen co-ords
	::POINT screenPoint = {0, 0};
	::GetCursorPos(&screenPoint);

	// Where cursor is in client area co-rds and clipped against client area
	::POINT clientPoint = screenPoint;
	clientPoint.x = min(max(clientScreenRect.left, clientPoint.x), clientScreenRect.right);
	clientPoint.y = min(max(clientScreenRect.top,  clientPoint.y), clientScreenRect.bottom);
	::ScreenToClient(applicationWindow, &clientPoint);

	if (lastMousePoint.x != clientPoint.x || lastMousePoint.y != clientPoint.y)
	{
		application->setMousePosition(seanw::math::Vector2d<float>(static_cast<float>(clientPoint.x), static_cast<float>(clientPoint.y)));

		if (::PtInRect(&clientScreenRect, screenPoint))
		{
			if (!mouseInWindow)
			{
				application->mouseEntered();
				mouseInWindow = true;
			}
		}
		else
		{
			if (mouseInWindow)
			{
				application->mouseExited();
				mouseInWindow = false;
			}
		}

		if (lmbDown)
		{
			application->mouseLeftDragged();
		}
		else if (rmbDown)
		{
			application->mouseRightDragged();
		}
		else
		{
			application->mouseMoved();
		}

		lastMousePoint = clientPoint;
	}
}

/** Creates application window */
void initWindow()
{
	::WNDCLASS windowClass;
	
	windowClass.style         = CS_HREDRAW | CS_VREDRAW;
	windowClass.lpfnWndProc   = windowProcedure;
	windowClass.cbClsExtra    = 0;
	windowClass.cbWndExtra    = 0;
	windowClass.hInstance     = applicationInstance;
	windowClass.hIcon         = ::LoadIcon(applicationInstance, MAKEINTRESOURCE(IDI_MAIN_ICON));
	windowClass.hCursor       = ::LoadCursor(0, IDC_CROSS);
	windowClass.hbrBackground = 0;
	windowClass.lpszMenuName  = 0;
	windowClass.lpszClassName = WINDOW_CLASS_NAME;
	
	if (!::RegisterClass(&windowClass))
	{
		throw std::runtime_error(seanw::misc::getErrorString(__FILE__, __LINE__, "Could not register window class.").c_str());
	}

	applicationWindow = ::CreateWindowEx(0,
	                                     WINDOW_CLASS_NAME,
	                                     APPLICATION_NAME,
	                                     WS_VISIBLE | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
	                                     CW_USEDEFAULT,
	                                     CW_USEDEFAULT,
	                                     WINDOW_WIDTH,
	                                     WINDOW_HEIGHT,
	                                     0,
	                                     ::LoadMenu(applicationInstance, MAKEINTRESOURCE(IDR_MAIN_MENU)),
	                                     applicationInstance,
	                                     0);

	if (applicationWindow == 0)
	{
		throw std::runtime_error(seanw::misc::getErrorString(__FILE__, __LINE__, "Could not create window.").c_str());
	}

	::GetClientRect(applicationWindow, &clientRect);
	restartApplication();

	::HDC dc = ::GetDC(applicationWindow);
	offscreen   = ::CreateCompatibleBitmap(dc, clientRect.right, clientRect.bottom);
	offscreenDc = ::CreateCompatibleDC(dc);
	::ReleaseDC(applicationWindow, dc);
}

/** Perform cleanup on window being destroyed. */
void destroyWindow()
{
	if (offscreenDc != 0)
	{
		::DeleteDC(offscreenDc);
	}

	if (offscreen != 0)
	{
		::DeleteObject(offscreen);
	}
}

/** Updates and draws application. */
void updateApplication()
{
	seanw::timer::Timer timer;

	updateMouse();
	RedrawWindow(applicationWindow, 0, 0, RDW_NOERASE | RDW_INVALIDATE | RDW_UPDATENOW); 

	// Slow down application if it is reaching maximum frames per second to stop it hogging the computer and locking it up
	timer.waitUntilTimeElapsed(1.0 / MAX_FPS);
}

/** Runs Sandbox application
 *  @return exit value from Win32 WM_QUIT message
 */
int runApplication()
{
	try
	{
		initWindow();

		for (;;)
		{
			::MSG message;

			while (::PeekMessage(&message, 0, 0, 0, PM_REMOVE))
			{
				if (message.message == WM_QUIT)
				{
					destroyWindow();
					
					return message.wParam;
				}

				::TranslateMessage(&message);
				::DispatchMessage(&message);
			}

			if (!::IsIconic(applicationWindow))
			{
				updateApplication();
			}
			else
			{
				// Wait for a message
				::WaitMessage();
			}
		}
		
	}
	catch (...)
	{
		destroyWindow();
		throw;
	}
}

/** Out of memory handler which immediately throws a std::bad_alloc exception. */
int badAllocHandler(::size_t)
{
	throw std::bad_alloc(seanw::misc::getErrorString(__FILE__, __LINE__, "Out of memory.").c_str());
}

void showFatalError(const char* errorMessage)
{
	std::ostringstream message;

	message << "A fatal error has occured. The application will quit after you press the OK button.\n\nError message:\n"
	        << errorMessage;

	::MessageBox(0, message.str().c_str(), "Fatal error", MB_OK | MB_ICONERROR | MB_SYSTEMMODAL);
}

}

int WINAPI WinMain(::HINSTANCE instance, ::HINSTANCE, ::LPSTR, int)
{
	try
	{
		// MSVC++ 6 fix to throw exception when memory exhaustion occurs
		::_set_new_handler(badAllocHandler);
		
		#ifdef _DEBUG
			// Enable memory leak checking
			::_CrtSetDbgFlag(::_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_LEAK_CHECK_DF);
			
			// Create console window that uses standard input and output streams, for debugging use
			::AllocConsole();
			::SetConsoleTitle("Debug console"); 
			::freopen("CONIN$",  "rt", stdin);
			::freopen("CONOUT$", "wt", stdout);
		#endif

		applicationInstance = instance;

		return runApplication();
	}
	catch (std::exception& e)
	{
		showFatalError((e.what()[0] == 0) ? "Unknown std::exception object thrown." : e.what());

		// Rethrow exception in debug build so debugger can deal with it
		#ifdef _DEBUG
			throw;
		#endif
	}
	catch (...)
	{
		// Show fatal error
		showFatalError("Unknown exception thrown.");
	
		// Rethrow exception in debug build so debugger can deal with it
		#ifdef _DEBUG
			throw;
		#endif
	}

	return 0;
}

#endif