// ====================================================
//
// Copyright (c) 2002 Sean Wilson. All Rights Reserved.
//
// ====================================================

#include "particle.h"

#include "misc.h"

namespace seanw
{
namespace sandbox
{

Particle::Particle(math::Vector2d<float> initPosition, float initMass) :
	position(initPosition),
	mass    (initMass),
	fixed   (false)
{
	if (mass <= 0)
	{
		throw std::out_of_range(misc::getErrorString(__FILE__, __LINE__, "Mass must be a positive number.").c_str());
	}
}

void Particle::write(std::ostream& outputStream) const
{
	outputStream << mass << "\n" << position << "\n" << velocity << "\n" << netForce << "\n" << fixed;
}

void Particle::read(std::istream& inputStream)
{
	inputStream >> mass >> position >> velocity >> netForce >> fixed;
}

void Particle::update(float deltaTime)
{
	if (!fixed)
	{
		math::Vector2d<float> acceleration = netForce / mass;

		velocity += acceleration * deltaTime;
	}

	netForce  = math::Vector2d<float>.ORIGIN;
	position += velocity * deltaTime;
}

void Particle::applyForce(math::Vector2d<float> force)
{
	netForce += force;
}

void Particle::applyDamping(float drag)
{
	applyForce(velocity * -drag);
}

void Particle::applyCollision(math::Vector2d<float> contactNormal, bool frictionOn, float coefficientOfRestitution, float staticSpeed, float staticFriction, float kineticFriction)
{
	math::Vector2d<float> normalVelocity  = contactNormal * contactNormal.dot(velocity);
	math::Vector2d<float> tangentVelocity = velocity      - normalVelocity;

	if (frictionOn)
	{
		float tangentSpeed = tangentVelocity.getLength();

		if (tangentSpeed > staticSpeed)
		{
			// Apply kinetic friction
			applyForce(tangentVelocity.normalise() * contactNormal.dot(netForce) * kineticFriction);
		}
		else
		{
			// Apply static friction
			applyForce(tangentVelocity.normalise() * contactNormal.dot(netForce) * staticFriction * (tangentSpeed / staticSpeed));
		}
	}

	// Calculate velocity after collision
	velocity = tangentVelocity - normalVelocity * coefficientOfRestitution;
}

void Particle::applyCollision(bool frictionOn, int boundsWidth, int boundsHeight, float coefficientOfRestitution, float staticSpeed, float staticFriction, float kineticFriction)
{
	if (position.getX() < 0)
	{
		position.setX(0);
		applyCollision(math::Vector2d<float>(+1, 0), frictionOn, coefficientOfRestitution, staticSpeed, staticFriction, kineticFriction);
	}
	else if (position.getX() > boundsWidth)
	{
		position.setX(static_cast<float>(boundsWidth));
		applyCollision(math::Vector2d<float>(-1, 0), frictionOn, coefficientOfRestitution, staticSpeed, staticFriction, kineticFriction);
	}

	if (position.getY() < 0)
	{
		position.setY(0);
		applyCollision(math::Vector2d<float>(0, +1), frictionOn, coefficientOfRestitution, staticSpeed, staticFriction, kineticFriction);
	}
	else if (position.getY() > boundsHeight)
	{
		position.setY(static_cast<float>(boundsHeight));
		applyCollision(math::Vector2d<float>(0, -1), frictionOn, coefficientOfRestitution, staticSpeed, staticFriction, kineticFriction);
	}
}

void Particle::stop()
{
	velocity = math::Vector2d<float>.ORIGIN;
	netForce = math::Vector2d<float>.ORIGIN;
}

void Particle::setFixed(bool newFixed)
{
	fixed = newFixed;

	if (fixed)
	{
		stop();
	}
}

std::ostream& operator<<(std::ostream& outputStream, const Particle& object)
{
	object.write(outputStream);

	return outputStream;
}

std::istream& operator>>(std::istream& inputStream, Particle& object)
{
	object.read(inputStream);

	return inputStream;
}

}
}