// ====================================================
//
// Copyright (c) 2002 Sean Wilson. All Rights Reserved.
//
// ====================================================

#ifndef HEADER_SEANW_SANDBOX_PARTICLE
#define HEADER_SEANW_SANDBOX_PARTICLE

#include "vector2d.h"

#include <fstream>

namespace seanw
{
namespace sandbox
{

/** Represents a point mass which forces can be applied to. */
class Particle
{
public:

	/** Constructs a new unfixed particle.
	 *  @param  initPosition      particle position.
	 *  @param  initMass          particle mass. Must be positive.
	 *  @throw  std::out_of_range if mass is non-positive.
	 */
	explicit Particle(math::Vector2d<float> initPosition = math::Vector2d<float>(), float initMass = 1);

	/** Writes particle to output stream.
	 *  @param outputStream stream to write to.
	 */
	void write(std::ostream& outputStream) const;

	/** Read particle from input stream.
	 *  @param inputStream stream to write to.
	 */	
	void read(std::istream& inputStream);

	/** Applies net force acting on particle and updates its position for time period using Euler integration. Net force is then set to zero.
	 *  @param deltaTime amount of time to advance in seconds.
	 */
	void update(float deltaTime);

	/** Adds force to net force being applied to particle.
	 *  @param force force to apply.
	 */
	void applyForce(math::Vector2d<float> force);

	/** Adds viscous damping force to net force being applied to particle.
	 *  @param drag drag coeffcient.
	 */
	void applyDamping(float drag);

	/** Changes particle's velocity and position if it has collided with rectangular boundary starting at the origin.
	 *  @param frictionOn               whether to apply friction if a collision occurs.
	 *  @param boundWidth               boundary width.
	 *  @param boundHeight              boundary height.
	 *  @param coefficientOfRestitution coefficent of restitution to apply if collision occurs.
	 *  @param staticSpeed              speed where particle is considered not moving.
	 *  @param staticFriction           coeffecient of static friction to apply if collision occurs.
	 *  @param kineticFriction          coeffecient of kinetic friction to apply if collision occurs.
	 */
	void applyCollision(bool frictionOn, int boundsWidth, int boundsHeight, float coefficientOfRestitution, float staticSpeed, float staticFriction, float kineticFriction);

	/** Changes particle's velocity because of collision with a surface.
     *  @param contactNormal            normalised normal to surface hit.
	 *  @param frictionOn               whether to apply friction.
	 *  @param coefficientOfRestitution coefficent of restitution.
	 *  @param staticSpeed              speed where particle is considered not moving.
	 *  @param staticFriction           coeffecient of static friction.
	 *  @param kineticFriction          coeffecient of kinetic friction.
	 */
	void applyCollision(math::Vector2d<float> contactNormal, bool frictionOn, float coefficientOfRestitution, float staticSpeed, float staticFriction, float kineticFriction);

	/** Sets net force acting on particle and its velocity to zero. */
	void stop();

	/** Sets fixed status of particle. Particle cannot move when fixed and will have no velocity.
	 *  @param newFixed new fixed status.
	 */
	void setFixed(bool newFixed);

	/** Sets position of particle.
	 *  @param newPostion new position.
	 */
	void setPosition(math::Vector2d<float> newPosition)
	{
		position = newPosition;
	}

	/** Sets velocity of particle.
	 *  @param newVelocity new velocity.
	 */
	void setVelocity(math::Vector2d<float> newVelocity)
	{
		velocity = newVelocity;
	}

	/** Returns position of particle. */
	math::Vector2d<float> getPosition() const
	{
		return position;
	}

	/** Returns velocity of particle. */
	math::Vector2d<float> getVelocity() const
	{
		return velocity;
	}

	/** Returns fixed status of particle. */
	bool isFixed() const
	{
		return fixed;
	}

	/** Returns mass of particle. */
	float getMass() const
	{
		return mass;
	}

private:
	/** Mass of particle. */
	float mass;
	/** Position of particle. */
	math::Vector2d<float> position;
	/** Velocity of particle. */
	math::Vector2d<float> velocity;
	/** Net force acting on particle. */
	math::Vector2d<float> netForce;
	/** Indicates if particle can move. */
	bool fixed;
};

std::ostream& operator<<(std::ostream& outputStream, const Particle& object);
std::istream& operator>>(std::istream& inputStream, Particle& object);

}
}

#endif