// ====================================================
//
// Copyright (c) 2002 Sean Wilson. All Rights Reserved.
//
// ====================================================

#include "sandbox.h"

namespace seanw
{
namespace sandbox
{

// Win32 fix for std::min and std::max being defined as macros in "windows.h"
#ifndef _WIN32
using std::min;
using std::max;
#endif

const int                   Sandbox::PARTICLE_RADIUS            =  4;
const int                   Sandbox::PARTICLE_SELECT_DISTANCE   = 10;
const int                   Sandbox::SPRING_SELECT_DISTANCE     =  5;
const float                 Sandbox::THROW_TIME_THRESHOLD       =   0.100f;
const float                 Sandbox::UPDATE_STEP_SIZE           =   0.002f;
const float                 Sandbox::MAX_STEP_SIZE              =   0.050f;
const float                 Sandbox::PARTICLE_MASS              =   1.000f;
const float                 Sandbox::WORLD_DAMPING              =   0.050f;
const float                 Sandbox::SPRING_DAMPING             =  20.000f;
const float                 Sandbox::SPRING_STIFFNESS           = 350.000f;
const float                 Sandbox::STATIC_SPEED               =   0.010f;
const float                 Sandbox::STATIC_FRICTION            =   0.500f;
const float                 Sandbox::KINETIC_FRICTION	        =   0.300f;
const float                 Sandbox::COEFFICIENT_OF_RESTITUTION =   0.800f;
const math::Vector2d<float> Sandbox::GRAVITATIONAL_PULL         = math::Vector2d<float>(0, +800);

Sandbox::Sandbox(int initBoundsWidth, int initBoundsHeight) :
	stepSize        (0),
	gravityOn       (true),
	frictionOn      (true),
	wasFixed        (false),
	mouseInside     (false),
	mode            (MODE_BUILD),
	boundsWidth     (initBoundsWidth),
	boundsHeight    (initBoundsHeight),
	grabbedParticle (particleList.end()),
	selectedParticle(particleList.end()),
	pointedParticle (particleList.end()),
	pointedSpring   (springList.end())
{
}

void Sandbox::writeParticleList(std::ostream& outputStream) const
{
	outputStream << particleList.size() << "\n";

	typedef std::ostream_iterator<Particle, char, std::char_traits<char> > ParticleOstreamIterator;
	std::copy(particleList.begin(), particleList.end(), ParticleOstreamIterator(outputStream, "\n"));
}
	
void Sandbox::writeSpringList(std::ostream& outputStream) const
{
	outputStream << springList.size() << "\n";
	
	for (SpringList::const_iterator spring = springList.begin(); spring != springList.end(); ++spring)
	{
		(*spring).write(outputStream, particleList.begin());
		outputStream << "\n";
	}
}

void Sandbox::write(std::ostream& outputStream) const
{
	outputStream << boundsWidth << "\n" << boundsHeight << "\n";

	writeParticleList(outputStream);
	writeSpringList(outputStream);

	outputStream << gravityOn << "\n" << frictionOn;
}

void Sandbox::readParticleList(std::istream& inputStream)
{
	particleList.clear();
	
	int particleListSize = 0;
	inputStream >> particleListSize;

	for (int i = 0; i < particleListSize; ++i)
	{
		Particle particle;
		
		inputStream >> particle;
		particleList.push_back(particle);
	}
}

void Sandbox::readSpringList(std::istream& inputStream)
{
	springList.clear();

	int springListSize = 0;
	inputStream >> springListSize;

	for (int i = 0; i < springListSize; ++i)
	{
		Spring spring;

		spring.read(inputStream, particleList.begin());
		springList.push_back(spring);
	}
}

void Sandbox::read(std::istream& inputStream)
{
	inputStream >> boundsWidth >> boundsHeight;

	readParticleList(inputStream);
	readSpringList(inputStream);

	inputStream >> gravityOn >> frictionOn;

	grabbedParticle  = particleList.end();
	selectedParticle = particleList.end();
	mode             = MODE_BUILD;
}

void Sandbox::setMode(Mode newMode)
{
	mode             = newMode;
	selectedParticle = particleList.end();
}

void Sandbox::setMousePosition(math::Vector2d<float> newMousePosition)
{
	deltaMousePosition = newMousePosition - lastMousePosition;
	lastMousePosition  = mousePosition;
	mousePosition      = newMousePosition;		
}

Sandbox::SpringList::iterator Sandbox::getClosestSpring()
{
	float                closestDistance = 0;
	SpringList::iterator closestSpring   = springList.end();

	for (SpringList::iterator spring = springList.begin(); spring != springList.end(); ++spring)
	{
		// Diagram of how shortest distance from pointC to line segment pointA to pointB is found:
		//
		//     pointA
		//       |
		// pointD|------pointC
		//       |_|
		//       |  Right angle
		//       |
		//     pointB
		//
		const math::Vector2d<float>& pointA = (*spring).getParticle1Position();
		const math::Vector2d<float>& pointB = (*spring).getParticle2Position();
		const math::Vector2d<float>& pointC = mousePosition;

		float distanceAD = ((pointC.getX() - pointA.getX()) * (pointB.getX() - pointA.getX()) +
		                    (pointC.getY() - pointA.getY()) * (pointB.getY() - pointA.getY())) / (pointB - pointA).getSquaredLength();

		// If pointD is along ray from pointA to pointB
		if (distanceAD >= 0)
		{
			math::Vector2d<float> pointD = pointA + (pointB - pointA) * distanceAD;

			// If pointD is on line segment from point A to pointB
			if ((pointD - pointA).getSquaredLength() <= (pointB - pointA).getSquaredLength())
			{
				float distanceCD = (pointD - pointC).getLength();

				// If pointD is within spring selection distance and is closest found so far
				if (distanceCD <= SPRING_SELECT_DISTANCE && (closestSpring == springList.end() || distanceCD < closestDistance))
				{
					closestSpring   = spring;
					closestDistance = distanceCD;
				}
			}
		}
	}

	return closestSpring;
}

bool Sandbox::isSpringAttached(ParticleList::const_iterator particle1, ParticleList::const_iterator particle2) const
{
	for (SpringList::const_iterator spring = springList.begin(); spring != springList.end(); ++spring)
	{
		if ((*spring).isAttachedTo(particle1) && (*spring).isAttachedTo(particle2))
		{
			return true;
		}
	}

	return false;
}

void Sandbox::mouseLeftPressed()
{
	if (pointedParticle != particleList.end())
	{
		grabbedParticle = pointedParticle;
		wasFixed        = (*pointedParticle).isFixed();
		
		(*pointedParticle).setFixed(true);
		
		if (mode == MODE_GLUE)
		{
			wasFixed = !wasFixed;
		}
		else if (mode == MODE_PLAY)
		{
			(*grabbedParticle).setPosition(mousePosition);
		}
		else if (selectedParticle == pointedParticle)
		{
			selectedParticle = particleList.end();
		}
		else
		{
			if (selectedParticle != particleList.end() && !isSpringAttached(selectedParticle, pointedParticle))
			{
				addSpring(Spring(selectedParticle, pointedParticle, SPRING_STIFFNESS, SPRING_DAMPING));
			}

			selectedParticle = pointedParticle;
		}
	}
	else if (mode == MODE_BUILD)
	{
		ParticleList::iterator addedParticle = addParticle(Particle(mousePosition, PARTICLE_MASS));

		if (selectedParticle != particleList.end())
		{
			addSpring(Spring(selectedParticle, addedParticle, SPRING_STIFFNESS, SPRING_DAMPING));
		}

		selectedParticle = addedParticle;
		grabbedParticle  = addedParticle;
		wasFixed         = false;
	}
}

void Sandbox::mouseRightPressed()
{
	if (mode == MODE_BUILD && grabbedParticle == particleList.end())
	{
		if (pointedSpring != springList.end())
		{
			springList.erase(pointedSpring);
			pointedSpring = springList.end();
		}
		
		if (pointedParticle != particleList.end())
		{
			removeParticle(pointedParticle);
			selectedParticle = particleList.end();
		}
	}
}

void Sandbox::mouseRightReleased()
{
}

void Sandbox::mouseRightDragged()
{
}

void Sandbox::mouseLeftDragged()
{
	lastDragged.reset();
	
	if (grabbedParticle != particleList.end())
	{
		(*grabbedParticle).setPosition(mousePosition);
	}
}

void Sandbox::mouseMoved()
{
}

void Sandbox::mouseLeftReleased()
{
	if (grabbedParticle != particleList.end())
	{
		(*grabbedParticle).setFixed(wasFixed);
		
		// If in play mode, grabbed particle was not originally fixed and mouse drag happen soon enough to be considered throwing
		if (mode == MODE_PLAY && !wasFixed && lastDragged.getTimeElapsed() < THROW_TIME_THRESHOLD)
		{
			(*grabbedParticle).setVelocity(deltaMousePosition / static_cast<float>(lastDragged.getTimeElapsed()));
		}

		grabbedParticle = particleList.end();
	}
}

void Sandbox::mouseExited()
{
	mouseInside = false;	
}

void Sandbox::mouseEntered()
{
	mouseInside = true;
}

Sandbox::ParticleList::iterator Sandbox::addParticle(const Particle& particle)
{
	particleList.push_back(particle);

	ParticleList::iterator back(particleList.end());
	std::advance(back, -1);
	return back;
}

Sandbox::SpringList::iterator Sandbox::addSpring(const Spring& spring)
{
	springList.push_back(spring);
	
	SpringList::iterator back(springList.end());
	std::advance(back, -1);
	return back;
}

void Sandbox::removeParticle(ParticleList::iterator particle)
{
	for (SpringList::iterator spring = springList.begin(); spring != springList.end();)
	{
		if ((*spring).isAttachedTo(particle))
		{
			spring = springList.erase(spring);
		}
		else
		{
			++spring;
		}
	}

	particleList.erase(particle);
}

void Sandbox::update(float deltaTime)
{
	for (SpringList::iterator spring = springList.begin(); spring != springList.end(); ++spring)
	{
		(*spring).update();
	}

	for (ParticleList::iterator particle = particleList.begin(); particle != particleList.end(); ++particle)
	{
		if (gravityOn)
		{
			(*particle).applyForce(GRAVITATIONAL_PULL * (*particle).getMass());
		}

		(*particle).applyDamping(WORLD_DAMPING);
		(*particle).applyCollision(frictionOn, boundsWidth, boundsHeight, COEFFICIENT_OF_RESTITUTION, STATIC_SPEED, STATIC_FRICTION, KINETIC_FRICTION);
		(*particle).update(deltaTime);
	}
}

/** Functor that finds closest particle to a position. */
class ClosestParticle
{
public:
	ClosestParticle(math::Vector2d<float> initPosition) : position(initPosition) { }

	bool operator() (Particle particle1, Particle particle2) const
	{
		return (position - particle1.getPosition()).getSquaredLength() < (position - particle2.getPosition()).getSquaredLength();
	}

private:
	math::Vector2d<float> position;
};

void Sandbox::update()
{
	// Clip time step size if it has grown too large
	stepSize = min(static_cast<float>(stepTimer.getTimeElapsed()), MAX_STEP_SIZE);
	
	stepTimer.reset();

	if (mouseInside)
	{
		pointedParticle = std::min_element(particleList.begin(), particleList.end(), ClosestParticle(mousePosition));

		// Particle not pointed to if not in play mode and mouse is not is selection distance
		if (mode != MODE_PLAY && (mousePosition - (*pointedParticle).getPosition()).getLength() > PARTICLE_SELECT_DISTANCE)
		{
			pointedParticle = particleList.end();
		}

		if (pointedParticle == particleList.end())
		{
			pointedSpring = getClosestSpring();
		}
		else
		{
			pointedSpring = springList.end();
		}
	}
	else
	{
		pointedParticle = particleList.end();
		pointedSpring   = springList.end();
	}

	if (mode == MODE_PLAY)
	{
		// Calculate how many substeps to for updating so different computer speeds can have different FPS but similar integration accuracy
		int numSubSteps = static_cast<int>(stepSize / UPDATE_STEP_SIZE);

		for (int i = 0; i < numSubSteps; ++i)
		{
			update(stepSize / numSubSteps);
		}
	}
}

void Sandbox::paintParticle(graphics::Graphics& graphics, math::Vector2d<float> position, bool fixed, bool highlighted)
{
	const graphics::Color& color = highlighted ? graphics::Color.BLUE : graphics::Color.BLACK;

	if (fixed)
	{
		graphics.setColor(graphics::Color.WHITE);
		graphics.fillCircle(static_cast<int>(position.getX()), static_cast<int>(position.getY()), PARTICLE_RADIUS);

		graphics.setColor(color);
		graphics.drawCircle(static_cast<int>(position.getX()), static_cast<int>(position.getY()), PARTICLE_RADIUS);
	}
	else
	{
		graphics.setColor(color);
		graphics.fillCircle(static_cast<int>(position.getX()), static_cast<int>(position.getY()), PARTICLE_RADIUS);
	}
}

void Sandbox::paintSpring(graphics::Graphics& graphics, math::Vector2d<float> position1, math::Vector2d<float> position2, bool highlighted)
{
	graphics.setColor(highlighted ? graphics::Color.BLUE : graphics::Color.GREY);

	graphics.drawLine(static_cast<int>(position1.getX()), static_cast<int>(position1.getY()),
	                  static_cast<int>(position2.getX()), static_cast<int>(position2.getY()));
}

void Sandbox::paintSprings(graphics::Graphics& graphics)
{
	for (SpringList::const_iterator spring = springList.begin(); spring != springList.end(); ++spring)
	{
		bool highlighted = mode == MODE_BUILD && spring == pointedSpring;
		paintSpring(graphics, (*spring).getParticle1Position(), (*spring).getParticle2Position(), highlighted);
	}

	if (mouseInside && grabbedParticle == particleList.end() && selectedParticle != particleList.end())
	{
		paintSpring(graphics, mousePosition, (*selectedParticle).getPosition(), false);
	}
}

void Sandbox::paintParticles(graphics::Graphics& graphics)
{
	ParticleList::iterator highlightedParticle = (grabbedParticle != particleList.end()) ? grabbedParticle : pointedParticle;

	for (ParticleList::iterator particle = particleList.begin(); particle != particleList.end(); ++particle)
	{
		bool fixed       = (grabbedParticle == particle           ) ? wasFixed : (*particle).isFixed();
		bool highlighted = (particle        == highlightedParticle);
		paintParticle(graphics, (*particle).getPosition(), fixed, highlighted);
	}
}

void Sandbox::paint(graphics::Graphics& graphics)
{
	update();
	
	graphics.setColor(graphics::Color.WHITE);
	graphics.fillRectangle(0, 0, boundsWidth, boundsHeight);

	paintSprings(graphics);
	paintParticles(graphics);
}

std::ostream& operator<<(std::ostream& outputStream, const Sandbox& object)
{
	object.write(outputStream);

	return outputStream;
}

std::istream& operator>>(std::istream& inputStream, Sandbox& object)
{
	object.read(inputStream);

	return inputStream;
}

}
}
