// ====================================================
//
// Copyright (c) 2002 Sean Wilson. All Rights Reserved.
//
// ====================================================

#ifndef HEADER_SEANW_SANDBOX_SANDBOX
#define HEADER_SEANW_SANDBOX_SANDBOX

#include "vector2d.h"
#include "graphics.h"
#include "particle.h"
#include "spring.h"
#include "timer.h"

#include <algorithm>
#include <iostream>
#include <list>

namespace seanw
{
namespace sandbox
{

/** A spring and mass world displayed on a graphics context that can be interacted with through mouse clicks. */
class Sandbox
{
public:
	/** Current manipulation mode. */
	enum Mode
	{
		/** Play mode ID. */
		MODE_PLAY,
		/** Build mode ID. */
		MODE_BUILD,
		/** Glue mode ID. */
		MODE_GLUE
	};

	/** Constructs new Sandbox.
	 *  @param initBoundsWidth  width of Sandbox graphics context.
	 *  @param initBoundsHeight height of Sandbox graphics context.
	 */
	Sandbox(int initBoundsWidth, int initBoundsHeight);

	/** Writes Sandbox to output stream.
	 *  @param outputStream stream to write to.
	 */
	void write(std::ostream& outputStream) const;

	/** Read Sandbox from input stream.
	 *  @param inputStream stream to write to.
	 */	
	void read(std::istream& inputStream);

	/** Call when left mouse button is pressed on graphics context. */
	void mouseLeftPressed();
	
	/** Call when left mouse button is released after being clicked on graphics context. */
	void mouseLeftReleased();

	/** Call when mouse dragged on graphics context using left mouse button. */
	void mouseLeftDragged();

	/** Call when right mouse button is pressed on graphics context. */
	void mouseRightPressed();
	
	/** Call when right mouse button is released after being clicked on graphics context. */
	void mouseRightReleased();

	/** Call when mouse dragged on graphics context using right mouse button. */
	void mouseRightDragged();
	
	/** Call when mouse is moved and not dragged over graphics context. */
	void mouseMoved();
	
	/** Call when mouse exits graphics context. */
	void mouseExited();
	
	/** Call when mouse enters graphics context. */
	void mouseEntered();

	/** Paints Sandbox to graphics context.
	 *  @param graphics graphics context to paint to.
	 */
	void paint(graphics::Graphics& graphics);

	/** Returns whether gravity is on. */
	bool isGravityOn() const
	{
		return gravityOn;
	}

	/** Returns whether friction is on. */
	bool isFrictionOn() const
	{
		return frictionOn;
	}

	/** Returns current manipulation mode. */
	Mode getMode() const
	{
		return mode;
	}

	/** Used to set current position of mouse cursor over graphics context. Should be called before any mouse event method.
	 *  @param newMousePosition current mouse position.
	 */
	void setMousePosition(math::Vector2d<float> newMousePosition);

	/** Sets current manipulation mode. */
	void setMode(Mode newMode);

	/** Sets whether gravity is on.
	 *  @param newGravityOn new gravity setting.
	 */
	void setGravity(bool newGravityOn)
	{
		gravityOn = newGravityOn;
	}

	/** Sets whether friction is on.
	 *  @param newFrictionOn new frction setting.
	 */
	void setFriction(bool newFrictionOn)
	{
		frictionOn = newFrictionOn;
	}

private:
	/** List of particles. */
	typedef std::list<Particle> ParticleList;
	/** List of springs. */
	typedef std::list<Spring> SpringList;

	/** Read particle list from input stream.
	 *  @param inputStream stream to write to.
	 */	
	void readParticleList(std::istream& inputStream);

	/** Read spring list from input stream.
	 *  @param inputStream stream to write to.
	 */	
	void readSpringList(std::istream& inputStream);

	/** Writes particle list to output stream.
	 *  @param outputStream stream to write to.
	 */
	void writeParticleList(std::ostream& outputStream) const;

	/** Writes spring list to output stream.
	 *  @param outputStream stream to write to.
	 */
	void writeSpringList(std::ostream& outputStream) const;

	/** Paints a particle to graphics context.
	 *  @param graphics    graphics context to paint to.
	 *  @param position    center of particle.
	 *  @param fixed       whether particle should be drawn as if it is fixed.
	 *  @param highlighted whether particle should be drawn as if it is highlighted.
	 */
	void paintParticle(graphics::Graphics& graphics, math::Vector2d<float> position, bool fixed, bool highlighted);
	
	/** Paints a spring to graphics context.
	 *  @param graphics    graphics context to paint to.
	 *  @param position1   spring start position.
	 *  @param position2   spring end position.
	 *  @param highlighted whether spring should be drawn as if it is highlighted.
	 */
	void paintSpring(graphics::Graphics& graphics, math::Vector2d<float> position1, math::Vector2d<float> position2, bool highlighted);

	/** Paints all springs to graphics context.
	 *  @param graphics graphics context to paint to.
	 */
	void paintSprings(graphics::Graphics& graphics);
	
	/** Paints all particles to graphics context.
	 *  @param graphics graphics context to paint to.
	 */
	void paintParticles(graphics::Graphics& graphics);
	
	/** Adds a new spring.
	 *  @param  spring spring to add.
	 *  @return        iterator to added spring.
	 */
	SpringList::iterator addSpring(const Spring& spring);

	/** Adds a new particle.
	 *  @param  particle particle to add.
	 *  @return          iterator to added particle.
	 */
	ParticleList::iterator addParticle(const Particle& particle);
	
	/** Returns iterator to spring whose line is closest to mouse position and is within the spring selection distance. A spring may not be found. */
	SpringList::iterator Sandbox::getClosestSpring();

	/** Update Sandbox logic. */
	void update();
	
	/** Removes particle and all springs attached to it.
	 *  @param particle iterator to particle to be remove.
	 */
	void removeParticle(ParticleList::iterator particle);

	/** Returns if a spring is attached to two particles.
	 *  @param particleIndex1 iterator to first particle.
	 *  @param particleIndex2 iterator to second particle.
	 *  @return               whether they are attached by a spring.
	 */
	bool isSpringAttached(ParticleList::const_iterator particle1, ParticleList::const_iterator particle2) const;

	/** Updates all springs and particles.
	 *  @param deltaTime update time in seconds.
	 */
	void update(float deltaTime);

	/** Default mass of particle. */
	static const float PARTICLE_MASS ;
	/** Default world damping. */
	static const float WORLD_DAMPING;
	/** Default spring damping. */
	static const float SPRING_DAMPING;
	/** Default spring stiffness. */
	static const float SPRING_STIFFNESS;
	/** Default velocity when particle is considered not moving. */
	static const float STATIC_SPEED;
	/** Default static friction. */
	static const float STATIC_FRICTION;
	/** Default kinetic friction. */
	static const float KINETIC_FRICTION;
	/** Default coeffecient of restitution. */
	static const float COEFFICIENT_OF_RESTITUTION;
	/** Default gravitation pull vector. */
	static const math::Vector2d<float> GRAVITATIONAL_PULL;
	
	/** Radius of particle circle. */
	static const int PARTICLE_RADIUS;
	/** Radius of circle particle where it can be considered to be pointed at. */
	static const int PARTICLE_SELECT_DISTANCE;
	/** Distance from spring line where it can be considered to be pointed at. */
	static const int SPRING_SELECT_DISTANCE;
	/** Maximum time in seconds between last mouse drag and button release to cause particle to be thrown. */
	static const float THROW_TIME_THRESHOLD;

	/** Time in seconds each spring and particle updates are to be done in. */
	static const float UPDATE_STEP_SIZE;
	/** Maximum step size that can be taken in seconds. */
	static const float MAX_STEP_SIZE;

	/** List of particles in world. */
	ParticleList particleList;
	/** List of springs in world. */
	SpringList springList;

	/** Spring that is currently being pointed at. */
	SpringList::iterator pointedSpring;
	/** Particle that is currently being pointed at. */
	ParticleList::iterator pointedParticle;
	/** Particle being held by mouse. */
	ParticleList::iterator grabbedParticle;
	/** Particle selected by mouse to have spring attached to. */
	ParticleList::iterator selectedParticle;
	/** Fixed status before a particle was grabbed. */
	bool wasFixed;
	
	/** Timer for timing how large update steps are. */
	seanw::timer::Timer stepTimer;
	/** Time it takes for Sandbox to be updated and drawn. */
	float stepSize;
	
	/** Last mouse position in graphics context. */
	math::Vector2d<float> mousePosition;
	/** Whether cursor is inside graphics context. */
	bool mouseInside;
	/** Last recorded mouse position. */
	math::Vector2d<float> lastMousePosition;
	/** Mouse displacement between last recorded positions. */
	math::Vector2d<float> deltaMousePosition;
	/** Stores time last mouse drag event was logged in world. */
	timer::Timer lastDragged;

	/** Width of graphics context. */
	int boundsWidth;
	/** Height of graphics context. */
	int boundsHeight;

	/** Current manipulation mode. */
	Mode mode;
	/** Whether gravity is on. */
	bool gravityOn;
	/** Whether friction is on. */
	bool frictionOn;
};

std::ostream& operator<<(std::ostream& outputStream, const Sandbox& object);
std::istream& operator>>(std::istream& inputStream, Sandbox& object);

}
}

#endif