// ====================================================
//
// Copyright (c) 2002 Sean Wilson. All Rights Reserved.
//
// ====================================================

#ifndef HEADER_SEANW_SANDBOX_SPRING
#define HEADER_SEANW_SANDBOX_SPRING

#include "misc.h"
#include "vector2d.h"
#include "particle.h"

#include <iostream>
#include <list>

namespace seanw
{
namespace sandbox
{

/** Represents a spring which is attached to two particles and can apply forces to them. */
class Spring
{
public:
	/** Constructs a new spring at its natural rest length.
	 *  @param initParticle1     iterator to first particle to be attached to. This must be a valid particle.
	 *  @param initParticle2     iterator to second particle to be attached to. This must be a valid particle.
	 *  @param initStiffness     spring's stiffness/Hooke's constant. Must be non-negative.
	 *  @param initDamping       spring's damping constant. Must be non-negative.
	 *  @throw std::out_of_range if stiffness or damping constants are not valid.    
	 */
	Spring(std::list<Particle>::iterator initParticle1, std::list<Particle>::iterator initParticle2, float initStiffness, float initDamping);
	
	/** Default spring contructor. Implemented for STL container construction only and should not be used otherwise. */
	Spring();

	/** Writes spring to output stream.
	 *  @param outputStream stream to write to.
	 *  @param listBegin    list containing both attached particles start iterator.
	 */
	void write(std::ostream& outputStream, std::list<Particle>::const_iterator listBegin) const;

	/** Read spring from input stream.
	 *  @param inputStream stream to write to.
	 *  @param listBegin   list containing both attached particles start iterator. List should have all particles in the same order when the spring was saved.
	 */	
	void read(std::istream& inputStream, std::list<Particle>::iterator listBegin);

	/** Returns whether a particle is attached to this spring.
	 *  @param  testParticle iterator to particle to check.
	 *  @return              true if particle is attached to this spring.
	 */
	bool isAttachedTo(std::list<Particle>::const_iterator testParticle) const;

	/** Applies forces to attached particles. */
	void update();

	/** Returns first attached particle's position. */
	math::Vector2d<float> getParticle1Position() const;

	/** Returns second attached particle's position. */
	math::Vector2d<float> getParticle2Position() const;

private:
	/** Iterator to first particle spring is attached to. */
	std::list<Particle>::iterator particle1;
	/** Iterator to second particle spring is attached to. */
	std::list<Particle>::iterator particle2;
	/** Spring damping constant. */
	float damping;
	/** Spring stiffness constant. */
	float stiffness;
	/** Spring rest length. */
	float restLength;
};

}
}

#endif