// ====================================================
//
// Copyright (c) 2002 Sean Wilson. All Rights Reserved.
//
// ====================================================

#ifndef HEADER_SEANW_MATH_VECTOR2D
#define HEADER_SEANW_MATH_VECTOR2D

#include <cmath>
#include <iostream>

// Win32 fix for std::sqrt, std::sin and std::cos not being place in std namespace.
#ifndef _WIN32
using std::sqrt;
using std::sin;
using std::cos;
#endif

namespace seanw
{
namespace math
{

/** Represents a 2D vector. */
template <class T> class Vector2d
{
public:
	/** Constructs a vector.
	 *  @param initX x-axis position.
	 *  @param initY y-axis position.
	 */
	explicit Vector2d(T initX = T(), T initY = T()) : x(initX), y(initY)
	{
	}

	/** Writes vector to output stream.
	 *  @param outputStream stream to write to.
	 */
	void write(std::ostream& outputStream) const
	{
		outputStream << x << "\n" << y;
	}

	/** Reads vector from input stream.
	 *  @param inputStream stream to write to.
	 */	
	void read(std::istream& inputStream)
	{
		inputStream >> x >> y;
	}

	/** Constructs a vector that points from a start to an end position i.e. (end - start).
	 *  @param start start vector.
	 *  @param end   end vector.
	 */
	Vector2d(const Vector2d& start, const Vector2d& end)
	{
		*this = end - start;
	}

	/** Returns addition of two vectors.
	 *  @param  rhs vector to add.
	 *  @return     (this + rhs).
	 */
	Vector2d operator+(const Vector2d& rhs) const
	{
		return Vector2d(x + rhs.x, y + rhs.y);
	}

	/** Returns unaltered vector. */
	Vector2d operator+() const
	{
		return *this;
	}

	/** Returns subtraction of two vectors.
	 *  @param  rhs vector to subtract.
	 *  @return     (this - rhs).
	 */
	Vector2d operator-(const Vector2d& rhs) const
	{
		return Vector2d(x - rhs.x, y - rhs.y);
	}

	/** Returns reversed vector. */
	Vector2d operator-() const
	{
		return Vector2d(-x, -y);
	}

	/** Returns scaled vector.
	 *  @param  rhs scale factor.
	 *  @return     (this * rhs).
	 */
	Vector2d operator*(T rhs) const
	{
		return Vector2d(x * rhs, y * rhs);
	}

	/** Returns multiplication of two vectors.
	 *  @param  rhs vector to multiply by.
	 *  @return     (this * rhs).
	 */
	Vector2d operator*(const Vector2d& rhs) const
	{
		return Vector2d(x * rhs.x, y * rhs.y);
	}

	/** Returns scaled vector.
	 *  @param  rhs scale factor.
	 *  @return     (this / rhs).
	 */
	Vector2d operator/(T rhs) const
	{
		return Vector2d(x / rhs, y / rhs);
	}

	/** Returns division of two vectors.
	 *  @param  rhs vector to divide by.
	 *  @return     (this / rhs).
	 */
	Vector2d operator/(const Vector2d& rhs) const
	{
		return Vector2d(x / rhs.x, y / rhs.y);
	}

	/** Returns dot product of two vectors.
	 *  @return (this dot rhs).
	 */
	T dot(const Vector2d& rhs)
	{
		return x * rhs.x + y * rhs.y;
	}

	/** Changes vector by adding another to it.
	 *  @param  rhs vector to add.
	 *  @return     (this + rhs)
	 */
	Vector2d operator+=(const Vector2d& rhs)
	{
		return *this = *this + rhs;
	}

	/** Changes vector by subrtacting another from it.
	 *  @param  rhs vector to subtract.
	 *  @return     (this - rhs)
	 */
	Vector2d operator-=(const Vector2d& rhs)
	{
		return *this = *this - rhs;
	}

	/** Changes vector by scaling it.
	 *  @param  rhs scale factor.
	 *  @return       (this * scale)
	 */
	Vector2d operator*=(T rhs)
	{
		return *this = *this * rhs;
	}

	/** Changes vector by scaling it by another.
	 *  @param  rhs vector to scale by.
	 *  @return     (this * rhs)
	 */
	Vector2d operator*=(const Vector2d& rhs)
	{
		return *this = *this * rhs;
	}

	/** Changes vector by scaling it.
	 *  @param  rhs scale factor.
	 *  @return     (this / rhs)
	 */
	Vector2d operator/=(T rhs)
	{
		return *this = *this / rhs;
	}

	/** Changes vector by scaling it by another.
	 *  @param  rhs vector to add.
	 *  @return     (this / rhs)
	 */
	Vector2d operator/=(const Vector2d& rhs)
	{
		return *this = *this / rhs;
	}

	/** Returns length of vector. */
	T getLength() const
	{
		return static_cast<T>(sqrt(x * x + y * y));
	}

	/** Returns squared length of vector. */
	T getSquaredLength() const
	{
		return x * x + y * y;
	}

	/** Returns normalised vector. */
	Vector2d normalise() const
	{
		T length = getLength();

		return (length == 0) ? *this : Vector2d(x / length, y / length);
	}

	/** Returns vector rotated about z-axis around origin.
	 *  @param radians amount to radian to rotate by.
	 *  @return        rotated vector.
	 */
	Vector2d rotate(T radians)
	{
		return Vector2d(x * static_cast<T>(cos(radians)) - y * static_cast<T>(sin(radians)),
		                x * static_cast<T>(sin(radians)) + y * static_cast<T>(cos(radians)));
	}

	/** Returns x component of vector. */
	T getX() const
	{
		return x;
	}
	
	/** Returns y component of vector. */
	T getY() const
	{
		return y;
	}

	/** Sets x component of vector.
	 *  @param newX new value for x component.
	 */
	void setX(T newX)
	{
		x = newX;
	}

	/** Sets y component of vector.
	 *  @param newY new value for y component.
	 */
	void setY(T newY)
	{
		y = newY;
	}

	/** Origin vector. */
	static const Vector2d ORIGIN;

private:
	/** Vector x-axis position. */
	T x;
	/** Vector y-axis position. */
	T y;
};

template <class T> const Vector2d<T> Vector2d<T>::ORIGIN;

template <class T> Vector2d<T> operator*(float lhs, const Vector2d<T>& rhs)
{
	return rhs * lhs;
}

template <class T> Vector2d<T> operator/(float lhs, const Vector2d<T>& rhs)
{
	return Vector2d(lhs / rhs.getX(), lhs / rhs.getY());
}

template <class T> std::ostream& operator<<(std::ostream& outputStream, const Vector2d<T>& object)
{
	object.write(outputStream);

	return outputStream;
}

template <class T> std::istream& operator>>(std::istream& inputStream, Vector2d<T>& object)
{
	object.read(inputStream);

	return inputStream;
}

}
}

#endif